<!DOCTYPE html>
<html>
<head>
<!--<link href="style.css" rel="stylesheet" />-->
<style>
body {
  font-family: helvetica,serif;
  font-size: 14px;
}

#cy {
  width: 100%;
  height: 100%;
  position: absolute;
  left: 0;
  top: 0;
  z-index: 10;
}

h1 {
  opacity: 0.5;
  font-size: 1em;
}

.floating {
    float: right;
    position: absolute;
    right: 10px;
    top: 10px;
    z-index: 100;
}

.button {
    background-color: #555555;
    border: none;
    color: white;
    padding: 6px 12px;
    text-align: center;
    text-decoration: none;
    display: inline-block;
    font-size: 14px;
    margin: 4px 2px;
    cursor: pointer;
    border-radius: 8px;
    box-shadow: 0 4px 8px 0 rgba(0,0,0,0.2), 0 3px 10px 0 rgba(0,0,0,0.19);
}

</style>
<meta charset=utf-8 />
<meta name="viewport" content="user-scalable=no, initial-scale=1.0, minimum-scale=1.0, maximum-scale=1.0, minimal-ui">
<title>{{title}}</title>
<!--<script src="js/cytoscape.min.js"></script>-->
<!--<script src="js/dagre.min.js"></script>-->
<!--<script src="js/cytoscape-dagre.js"></script>-->

<script src="https://cdn.rawgit.com/cytoscape/cytoscape.js/v3.2.3/dist/cytoscape.min.js"></script>
<script src="https://cdn.rawgit.com/cpettitt/dagre/v0.7.4/dist/dagre.min.js"></script>
<script src="https://cdn.rawgit.com/cytoscape/cytoscape.js-dagre/1.5.0/cytoscape-dagre.js"></script>

<script src="http://ajax.googleapis.com/ajax/libs/jquery/1/jquery.min.js"></script>
<script src="http://cdnjs.cloudflare.com/ajax/libs/qtip2/2.2.0/jquery.qtip.min.js"></script>
<link href="http://cdnjs.cloudflare.com/ajax/libs/qtip2/2.2.0/jquery.qtip.min.css" rel="stylesheet" type="text/css" />
<script src="https://cdn.rawgit.com/cytoscape/cytoscape.js-qtip/2.7.0/cytoscape-qtip.js"></script>

</head>
<body>
<div id="cy"></div>
<div class="floating">
<a class="button" href="#" id="downloader" onclick="to_image()" download="{{title}}.jpg">Download (jpg)</a>
</div>
<!-- Load appplication code at the end to ensure DOM is loaded -->
<script type="text/javascript">{{graph}}</script>
</body>
</html>
